from ..common.log import log_current_exception
from ..sql import Sector, SectorGroup
from ..sql.types import short_cls
from ..sql.utils import add
from ..web.servlets.sector import UserDefined


def import_sector(record, old, new):
    if not sector_imported(record, new):
        record.info('Importing sector entries')
        try:
            with old.session_context() as old_s:
                with new.session_context() as new_s:
                    groups = copy_sector_group(record, old_s, old, new_s)
                    copy_sector(record, old_s, old, new_s, groups)
        except Exception as e:
            log_current_exception()
            record.warning(f'Aborting sector import: {e}')
    else:
        record.warning('Sector entries already imported')


def sector_imported(record, new):
    with new.session_context() as new_s:
        return bool(new_s.query(Sector).count())


def copy_sector_group(record, old_s, old, new_s):
    groups = {}
    sector_group = old.meta.tables['sector_group']
    for old_sector_group in old_s.query(sector_group).all():
        new_sector_group = match_sector_group(new_s, old_sector_group)
        if new_sector_group:
            groups[new_sector_group.title] = new_sector_group
        else:
            groups[new_sector_group.title] = \
                add(new_s, SectorGroup(srid=old_sector_group.srid, centre=old_sector_group.centre,
                                       radius=old_sector_group.radius, title=old_sector_group.title))
            record.info(f'Sector Group {old_sector_group.title}')
    return groups


def match_sector_group(new_s, old_sector_group):
    # we could do a bunch of things here, but this is sufficient to avoid the worst errors
    return new_s.query(SectorGroup). \
        filter(SectorGroup.title == old_sector_group.title,
               SectorGroup.srid == old_sector_group.srid).one_or_none()


def copy_sector(record, old_s, old, new_s, groups):
    sector = old.meta.tables['sector']
    sector_group = old.meta.tables['sector_group']
    for old_sector in old_s.query(sector).filter(sector.c.owner == short_cls(UserDefined)).all():
        old_sector_group = old_s.query(sector_group).filter(sector_group.c.id == old_sector.sector_group_id).one()
        new_sector_group = groups[old_sector_group.title]
        add(new_s, Sector(sector_group_id=new_sector_group.id, route=old_sector.route, title=old_sector.title,
                          owner=UserDefined, distance=old_sector.distance,
                          start=old_sector.start, finish=old_sector.finish, hull=old_sector.hull))
        record.info(f'Sector {old_sector.title}')




from logging import getLogger

from sqlalchemy.sql.functions import count

from ch2.commands.args import V, bootstrap_db
from ch2.common.args import m
from ch2.common.date import to_date
from ch2.config.profiles.acooke import acooke
from ch2.pipeline.calculate.summary import SummaryCalculator
from ch2.sql.tables.source import Source, Interval
from ch2.sql.tables.statistic import StatisticJournalText, StatisticJournal, StatisticJournalFloat, StatisticName, \
    StatisticJournalInteger, StatisticJournalType
from ch2.sql.tables.topic import DiaryTopicJournal, DiaryTopic
from ch2.sql.utils import add
from tests import LogTestCase, random_test_user

log = getLogger(__name__)


# the idea here is to test the new database schema with sources etc
# so we configure a database then load some data, calculate some stats,
# and see if everything works as expected.

class TestSources(LogTestCase):

    def test_sources(self):

        user = random_test_user()
        config = bootstrap_db(user, m(V), '5', configurator=acooke)

        with config.db.session_context() as s:

            # add a diary entry

            journal = add(s, DiaryTopicJournal(date='2018-09-29'))
            cache = journal.cache(s)
            diary = s.query(DiaryTopic).filter(DiaryTopic.title == 'Status').one()
            fields = diary.fields
            self.assertEqual(len(fields), 6, list(enumerate(map(str, fields))))
            self.assertEqual(fields[0].statistic_name.name, 'notes')
            self.assertEqual(fields[1].statistic_name.name, 'weight', str(fields[1]))
            statistics = [cache[field] for field in fields]
            for statistic in statistics:
                self.assertTrue(statistic.value is None, statistics)
            statistics[0].value = 'hello world'
            statistics[1].value = 64.5

        with config.db.session_context() as s:

            # check the diary entry was persisted

            journal = DiaryTopicJournal.get_or_add(s, '2018-09-29')
            cache = journal.cache(s)
            diary = s.query(DiaryTopic).filter(DiaryTopic.title == 'Status').one()
            fields = diary.fields
            self.assertEqual(len(fields), 6, list(enumerate(map(str, fields))))
            self.assertEqual(fields[0].statistic_name.name, 'notes')
            self.assertEqual(fields[1].statistic_name.name, 'weight', str(fields[1]))
            statistics = [cache[field] for field in fields]
            self.assertEqual(statistics[1].value, 64.5)
            self.assertEqual(statistics[1].type, StatisticJournalType.FLOAT)

        # generate summary stats

        SummaryCalculator(config, schedule='m').run()
        SummaryCalculator(config, schedule='y').run()

        with config.db.session_context() as s:

            # check the summary stats

            diary = s.query(DiaryTopic).filter(DiaryTopic.title == 'Status').one()
            weights = s.query(StatisticJournal).join(StatisticName). \
                           filter(StatisticName.owner == diary, StatisticName.name == 'weight'). \
                           order_by(StatisticJournal.time).all()
            self.assertEqual(len(weights), 2)
            self.assertEqual(weights[1].value, 64.5)
            self.assertEqual(len(weights[1].measures), 2, weights[1].measures)
            self.assertEqual(weights[1].measures[0].rank, 1)
            self.assertEqual(weights[1].measures[0].percentile, 100, weights[1].measures[0].percentile)
            n = s.query(count(StatisticJournalFloat.id)).scalar()
            self.assertEqual(n, 4, n)
            n = s.query(count(StatisticJournalInteger.id)).scalar()
            self.assertEqual(n, 6, n)
            m_avg = s.query(StatisticJournalFloat).join(StatisticName). \
                filter(StatisticName.name == 'avg-month-weight').one()
            self.assertEqual(m_avg.value, 64.5)
            y_avg = s.query(StatisticJournalFloat).join(StatisticName). \
                filter(StatisticName.name == 'avg-year-weight').one()
            self.assertEqual(y_avg.value, 64.5)
            month = s.query(Interval).filter(Interval.schedule == 'm').first()
            self.assertEqual(month.start, to_date('2018-09-01'), month.start)
            self.assertEqual(month.finish, to_date('2018-10-01'), month.finish)

        with config.db.session_context() as s:

            # delete the diary entry

            journal = DiaryTopicJournal.get_or_add(s, '2018-09-29')
            s.delete(journal)

        with config.db.session_context() as s:

            # check the delete cascade

            self.assertEqual(s.query(count(DiaryTopicJournal.id)).scalar(), 1)
            # this should be zero because the Intervals were automatically deleted
            # (well, now +1 because there's an original default weight)
            for source in s.query(Source).all():
                print(source)
            for journal in s.query(StatisticJournal).all():
                print(journal)
            self.assertEqual(s.query(count(Source.id)).scalar(), 36, list(map(str, s.query(Source).all())))  # constants
            self.assertEqual(s.query(count(StatisticJournalText.id)).scalar(), 13, s.query(count(StatisticJournalText.id)).scalar())
            self.assertEqual(s.query(count(StatisticJournal.id)).scalar(), 22, s.query(count(StatisticJournal.id)).scalar())



from contextlib import contextmanager
from logging import getLogger, DEBUG, INFO, WARNING
from os.path import join

from ..commands.args import LOG_DIR
from ..common.log import configure_log, log_current_exception, set_log_color
from ..common.names import UNDEF, COLOR

log = getLogger(__name__)


def make_log_from_args(args):
    from ..commands.args import LOG, COMMAND, VERBOSITY, PROGNAME, DEV
    name = args[LOG] if LOG in args and args[LOG] else (
            (args[COMMAND] if COMMAND in args and args[COMMAND] else PROGNAME) + f'.{LOG}')
    path = join(args._format_path(LOG_DIR), name)
    if args[VERBOSITY] is UNDEF:
        verbosity = 5 if args[DEV] else 4
    else:
        verbosity = args[VERBOSITY]
    configure_log('ch2', path, verbosity, {
        'bokeh': DEBUG,
        'ch2': DEBUG,
        'jupyter': DEBUG,
        'matplotlib': DEBUG,
        'sentinelsat': DEBUG,
        'sqlalchemy': WARNING,
        'tornado': DEBUG,
        'werkzeug': DEBUG,
        '__main__': DEBUG
    })
    set_log_color(args[COLOR])
    log.info(f'Logging to {path}')


class Record:

    def __init__(self, log):
        self._log = log
        self._warning = []
        self._info = []

    def warning(self, msg):
        self._log.warning(msg)
        self._warning.append(msg)

    def info(self, msg):
        self._log.info(msg)
        self._info.append(msg)

    def raise_(self, msg):
        self.warning(msg)
        raise Exception(msg)

    @contextmanager
    def record_exceptions(self, catch=False):
        try:
            yield
        except Exception as e:
            self.warning(e)
            if catch:
                log_current_exception()
            else:
                raise

    def json(self):
        return {'warning': self._warning,
                'info': self._info}



from ..climb import CLIMB_CNAME
from ..database import add_diary_topic, add_child_diary_topic, add_diary_topic_field, add_process, add_pipeline
from ..power import add_simple_power_estimate, add_kit_power_estimate, add_kit_power_model, POWER_MODEL_CNAME
from ..profile import WALK, SWIM, RUN, BIKE, Profile
from ...commands.args import DEFAULT
from ...common.names import TIME_ZERO
from ...diary.model import TYPE, EDIT
from ...lib import to_time, time_to_local_date
from ...names import Sports, simple_name, N
from ...pipeline.calculate import SectorCalculator, ElevationCalculator
from ...pipeline.calculate.climb import FindClimbCalculator
from ...pipeline.calculate.cluster import ClusterCalculator
from ...pipeline.calculate.power import PowerCalculator
from ...pipeline.calculate.sector import NewSectorCalculator
from ...pipeline.read.activity import ActivityReader
from ...sql import StatisticJournalType, StatisticName, DiaryTopic, DiaryTopicJournal, PipelineType
from ...sql.tables.sector import SectorGroup, DEFAULT_GROUP_RADIUS_KM
from ...sql.tables.statistic import STATISTIC_JOURNAL_CLASSES
from ...sql.types import short_cls
from ...sql.utils import add


def acooke(config):
    '''
## acooke

This extends the garmin configuration with:

* Diary entries that I need
* Additional activity groups selected on the kit used
* Power estimates

Unlikely to be useful to others, but works as an example of how you can extend the code yourself.
    '''
    ACooke(config).load()


ROAD = 'Road'
MTB = 'MTB'


class ACooke(Profile):

    def _load_diary_topics(self, s, c):
        super()._load_diary_topics(s, c)

        # add more diary entries specific to my needs

        injuries = add_diary_topic(s, 'Injuries', c)

        ms = add_child_diary_topic(s, injuries, 'Multiple Sclerosis', c)
        add_diary_topic_field(s, ms, 'MS Notes', c, StatisticJournalType.TEXT, model={TYPE: EDIT})
        add_child_diary_topic(s, ms, 'Betaferon', c,
                              schedule='2018-08-08/2d[1]')  # reminder to take meds on alternate days

        leg = add_child_diary_topic(s, injuries, 'Broken Femur LHS', c,
                                    schedule='d2018-03-11-2020-03-01')
        add_diary_topic_field(s, leg, 'Leg Notes', c, StatisticJournalType.TEXT, model={TYPE: EDIT})
        add_child_diary_topic(s, leg, 'Learn to manage tendon pain', c)  # aims added as child topics
        add_child_diary_topic(s, leg, 'Maintain fitness', c)
        add_child_diary_topic(s, leg, 'Visit UK', c, schedule='-2018-08-11')

    def _load_specific_activity_groups(self, s):
        super()._load_specific_activity_groups(s)

        # additional activity groups for different cycling activities

        self._load_activity_group(s, ROAD, 'Road cycling activities')
        self._load_activity_group(s, MTB, 'MTB cycling activities')

    def _load_sector_groups(self, s):
        # note lon, lat for centre
        SectorGroup.add(s, (-70.7, -33.4), DEFAULT_GROUP_RADIUS_KM, 'Santiago, Chile')

    def _sport_to_activity(self):

        # map the additional groups above based on kit use
        # (cotic and bowman are kit items added via kit commands)

        return {Sports.SPORT_CYCLING: {
                    N.KIT: {
                        'cotic': simple_name(MTB),
                        'bowman': simple_name(ROAD),
                        'cbutler': simple_name(ROAD),
                    },
                    DEFAULT: simple_name(BIKE),
                },
                Sports.SPORT_RUNNING: simple_name(RUN),
                Sports.SPORT_SWIMMING: simple_name(SWIM),
                Sports.SPORT_WALKING: simple_name(WALK)}

    def _sector_statistics(self, s, power_statistics=None):
        power_statistics = power_statistics or []
        add_process(s, ClusterCalculator, blocked_by=[ElevationCalculator],
                    owner_in=short_cls(ActivityReader))
        for activity_group in (ROAD, MTB):
            add_process(s, FindClimbCalculator, blocked_by=[ElevationCalculator],
                        owner_in=short_cls(ActivityReader), climb=CLIMB_CNAME,
                        activity_group=activity_group)
            add_process(s, SectorCalculator, blocked_by=[ClusterCalculator, FindClimbCalculator],
                        power_model=POWER_MODEL_CNAME, activity_group=activity_group)
        return power_statistics + [SectorCalculator]

    def _load_sector_pipeline(self, s):
        for activity_group in (ROAD, MTB):
            add_pipeline(s, NewSectorCalculator, PipelineType.SECTOR, activity_group=activity_group)

    def _load_power_statistics(self, s, simple=False):
        # add power estimates for the two bikes
        # (note that this comes after standard stats, but before summary, achievements, etc).
        if simple:
            for activity_group in (MTB, ROAD):
                activity_group = self._activity_groups[activity_group]
                add_simple_power_estimate(s, activity_group, 0.42, 0.0055, 12, 65)
        else:
            add_kit_power_estimate(s, (MTB, ROAD))
            for kit, activity_group, cda, crr, bike_weight in (('cotic', MTB, 0.42, 0.0055, 12),
                                                               ('bowman', ROAD, 0.42, 0.0055, 8),
                                                               ('cbutler', ROAD, 0.42, 0.0055, 12)):
                add_kit_power_model(s, kit, self._activity_groups[activity_group], cda, crr, bike_weight)
        return [PowerCalculator]  # additional blocker for activity stats

    def _post(self, s):
        # set a default weight for early power calculations
        weight = s.query(StatisticName).filter(StatisticName.name == 'Weight', StatisticName.owner == DiaryTopic).one()
        diary = add(s, DiaryTopicJournal(date=time_to_local_date(to_time(0.0))))
        add(s, STATISTIC_JOURNAL_CLASSES[weight.statistic_journal_type](
            value=65.0, time=TIME_ZERO, statistic_name=weight, source=diary))
        super()._post(s)



from logging import getLogger

from .names import VERSION, URI

log = getLogger(__name__)


class BaseConfig:

    def __init__(self, args, factory, db_version):
        self.args = args
        self.args[VERSION] = db_version
        self.__factory = factory
        self.__db = None

    @property
    def db(self):
        if not self.__db:
            self.__db = self.get_database()
        return self.__db

    def reset(self):
        self.__db = None

    def get_database(self, **kargs):
        # prefer kargs to _with so that passwd is not displayed
        args = self.args._with(**kargs)
        safe_uri = args._with(passwd='xxxxxx')._format(URI)
        log.debug(f'Connecting to {safe_uri}')
        uri = args._format(URI)
        return self.__factory(uri)

    def _with(self, **kargs):
        # may need to be over-written by subclasses
        args = self.args._with(**kargs)
        return BaseConfig(args, self.__factory, args[VERSION])




from .achievement import Achievement
from .activity import ActivityGroup, ActivityTimespan, ActivityJournal, ActivityBookmark
from .cluster import ClusterInputScratch, ClusterHull, ClusterFragmentScratch
from .constant import Constant
from .file import FileScan, FileHash
from .kit import KitGroup, KitItem, KitComponent, KitModel
from .nearby import ActivitySimilarity, ActivityNearby
from .pipeline import Pipeline, PipelineType
from .sector import SectorGroup, Sector, SectorClimb, SectorJournal, SectorType
from .source import Source, Interval, NoStatistics, Composite, CompositeComponent
from .statistic import StatisticName, StatisticJournalFloat, StatisticJournalText, StatisticJournalInteger, \
    StatisticJournalTimestamp, StatisticJournal, StatisticMeasure, StatisticJournalType
from .system import SystemConstant, Process
from .timestamp import Timestamp
from .topic import DiaryTopicJournal, DiaryTopic, DiaryTopicField, ActivityTopicJournal, ActivityTopic, \
    ActivityTopicField



import datetime as dt
from enum import IntEnum
from logging import getLogger
from random import choice
from string import ascii_letters
from unittest import TestCase

import pytz
from sqlalchemy import create_engine, Integer, Column, union, ForeignKey, UniqueConstraint, Text, \
    TypeDecorator, DateTime, Index
from sqlalchemy.orm import declarative_base, sessionmaker, relationship, backref

log = getLogger(__name__)


class SQLAlchemyTest(TestCase):

    def test_sqlalchemy(self):

        # using postgres with log_statement=all so that we can see the incorrect queries
        # (use a transient docker instance)

        dbname = ''.join(choice(ascii_letters) for _ in range(16)).lower()
        # https://stackoverflow.com/questions/6506578/how-to-create-a-new-database-using-sqlalchemy
        engine = create_engine('postgresql://postgres@localhost:5432/postgres')
        conn = engine.connect()
        conn.execute('commit')
        conn.execute(f'create database {dbname}')
        conn.close()

        engine = create_engine(f'postgresql://postgres@localhost:5432/{dbname}')
        Base = declarative_base()
        Session = sessionmaker(engine)

        class SourceType(IntEnum):
            SOURCE = 0
            ACTIVITY = 2
            ACTIVITY_TOPIC = 10

        class StatisticJournalType(IntEnum):
            STATISTIC = 0
            TIMESTAMP = 4

        class FileHash(Base):
            __tablename__ = 'file_hash'
            id = Column(Integer, primary_key=True)

        class Source(Base):
            __tablename__ = 'source'
            id = Column(Integer, primary_key=True)
            type = Column(Integer, nullable=False, index=True)
            __mapper_args__ = {
                'polymorphic_identity': SourceType.SOURCE,
                'polymorphic_on': type
            }

        class GroupedSource(Source):
            __abstract__ = True

        class ActivityJournal(GroupedSource):
            __tablename__ = 'activity_journal'
            id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
            file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False, index=True, unique=True)
            file_hash = relationship('FileHash', backref=backref('activity_journal', uselist=False))
            __mapper_args__ = {
                'polymorphic_identity': SourceType.ACTIVITY
            }

        class ActivityTopicJournal(GroupedSource):
            __tablename__ = 'activity_topic_journal'
            id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
            file_hash_id = Column(Integer, ForeignKey('file_hash.id'),
                                  nullable=False, index=True, unique=True)
            file_hash = relationship('FileHash', backref=backref('activity_topic_journal', uselist=False))
            __mapper_args__ = {
                'polymorphic_identity': SourceType.ACTIVITY_TOPIC
            }

        class StatisticName(Base):
            __tablename__ = 'statistic_name'
            id = Column(Integer, primary_key=True)
            name = Column(Text, nullable=False)

        class StatisticJournal(Base):
            __tablename__ = 'statistic_journal'
            id = Column(Integer, primary_key=True)
            type = Column(Integer, nullable=False, index=True)
            statistic_name_id = Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'), nullable=False)
            statistic_name = relationship('StatisticName')
            source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), nullable=False)
            source = relationship('Source')
            __mapper_args__ = {
                'polymorphic_identity': StatisticJournalType.STATISTIC,
                'polymorphic_on': 'type'
            }

        class StatisticJournalTimestamp(StatisticJournal):
            __tablename__ = 'statistic_journal_timestamp'
            id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
            value = Column(DateTime, nullable=False)
            __mapper_args__ = {
                'polymorphic_identity': StatisticJournalType.TIMESTAMP
            }

        Base.metadata.create_all(engine)

        def build_source_query(s, value):
            q = s.query(Source.id). \
                join(StatisticJournalTimestamp). \
                join(StatisticName). \
                filter(StatisticName.name.like('start')). \
                filter(StatisticJournalTimestamp.value > value)
            q_direct = s.query(ActivityJournal.id). \
                filter(ActivityJournal.id.in_(q.subquery().select()))
            q_via_topic = s.query(ActivityJournal.id). \
                join(FileHash). \
                join(ActivityTopicJournal). \
                filter(ActivityTopicJournal.id.in_(q.subquery().select()))
            constraints = union(q_direct, q_via_topic).subquery().select()
            return s.query(Source).filter(Source.id.in_(constraints))

        with Session() as s:
            build_source_query(s, dt.datetime(2020, 1, 1, 3, 0, tzinfo=pytz.UTC)).all()
            build_source_query(s, dt.datetime(2021, 1, 1, 3, 0, tzinfo=pytz.UTC)).all()




import datetime as dt
from logging import getLogger

from sqlalchemy.sql.functions import count

from ch2.commands.args import V, DEV, bootstrap_db
from ch2.common.args import mm, m
from ch2.config.profiles.default import default
from ch2.sql import Source
from ch2.sql.tables.timestamp import Timestamp
from ch2.sql.utils import add
from tests import LogTestCase, random_test_user

log = getLogger(__name__)


class TestTimestamp(LogTestCase):

    def test_set(self):
        user = random_test_user()
        bootstrap_db(user, m(V), '5')
        config = bootstrap_db(user, m(V), '5', mm(DEV), configurator=default)
        with config.db.session_context() as s:
            source = add(s, Source())
            n = s.query(count(Timestamp.id)).scalar()
            self.assertEqual(n, 0)
            Timestamp.set(s, TestTimestamp, source=source)
            n = s.query(count(Timestamp.id)).scalar()
            self.assertEqual(n, 1)
            t = s.query(Timestamp).filter(Timestamp.owner == TestTimestamp).one()
            self.assertAlmostEqual(t.time.timestamp(), dt.datetime.now().timestamp(), 1)

    def test_context(self):
        user = random_test_user()
        bootstrap_db(user, m(V), '5')
        config = bootstrap_db(user, m(V), '5', mm(DEV), configurator=default)
        with config.db.session_context() as s:
            with Timestamp(owner=TestTimestamp).on_success(s):
                n = s.query(count(Timestamp.id)).scalar()
                self.assertEqual(n, 0)
            n = s.query(count(Timestamp.id)).scalar()
            self.assertEqual(n, 1)

    def test_context_error(self):
        user = random_test_user()
        bootstrap_db(user, m(V), '5')
        config = bootstrap_db(user, m(V), '5', mm(DEV), configurator=default)
        with config.db.session_context() as s:
            try:
                with Timestamp(owner=TestTimestamp).on_success(s):
                    n = s.query(count(Timestamp.id)).scalar()
                    self.assertEqual(n, 0)
                    raise Exception('foo')
            except Exception as e:
                self.assertEqual(str(e), 'foo')  # for some weird reason assertRaisesRegex was not working
            n = s.query(count(Timestamp.id)).scalar()
            self.assertEqual(n, 0)




import datetime as dt

from .load import display_notebook
from ..common.date import format_date, time_to_local_time


def stringify(value):
    # this has auto-conversion of UTC to local times (since interface is public)
    if isinstance(value, str):
        return value
    elif isinstance(value, dt.datetime):
        return time_to_local_time(value)
    elif isinstance(value, dt.date):
        return format_date(value)
    else:
        return str(value)


def template(func):

    def wrapper(*args, config=None):
        if config:
            sargs = [stringify(arg) for arg in args]
            display_notebook(config, func, sargs)
        else:
            return func(*args)

    wrapper._original = func

    return wrapper




from logging import getLogger

from .utils import ProcessCalculator, ActivityJournalProcessCalculator
from ..pipeline import OwnerInMixin
from ..read.activity import ActivityReader
from ...common.date import local_time_to_time
from ...common.log import log_current_exception
from ...sql import StatisticJournal, Timestamp
from ...sql.tables.kit import expand_item

log = getLogger(__name__)


class KitCalculator(OwnerInMixin, ActivityJournalProcessCalculator):
    '''
    Convert `-D kit=XXX` statistic set on activities during import into kit usage entries.

    The kit statistic is also used directly by other code (eg to filter activities by kit).
    '''

    def _run_one(self, missed):
        start = local_time_to_time(missed)
        with self._config.db.session_context() as s:
            ajournal = self._get_source(s, start)
            with Timestamp(owner=self.owner_out, source=ajournal).on_success(s):
                kit = StatisticJournal.for_source(s, ajournal.id, ActivityReader.KIT, self.owner_in,
                                                  ajournal.activity_group)
                if kit:
                    log.debug(f'Read {kit.value} at {missed} / {ajournal.activity_group.name}')
                    for kit_name in kit.value.split(','):
                        try:
                            for kit_instance in expand_item(s, kit_name, ajournal.start):
                                kit_instance.add_use(s, ajournal.start, source=ajournal, owner=self.owner_out)
                                log.debug(f'Added usage for {kit_instance}')
                        except Exception as e:
                            log_current_exception()
                            log.warning(f'Could not add statistics for {kit_name}: {e}')
                else:
                    log.debug(f'No kit defined for this activity ({missed})')



from logging import getLogger
from math import sqrt

import pandas as pd
from geoalchemy2.shape import to_shape
from sqlalchemy import text, select

from .utils import ActivityJournalProcessCalculator, DataFrameCalculatorMixin
from ..pipeline import LoaderMixin
from ...common.geo import utm_srid
from ...common.math import is_nan
from ...data import Statistics
from ...data.activity import add_delta_azimuth
from ...data.elevation import smooth_elevation, add_gradient
from ...data.frame import present
from ...names import N, T, U
from ...sql import StatisticJournalType, ActivityJournal, StatisticJournal
from ...sql.types import linestringxyzm, linestringxym

log = getLogger(__name__)


class ElevationCalculator(LoaderMixin, DataFrameCalculatorMixin, ActivityJournalProcessCalculator):

    def __init__(self, *args, smooth=3, **kargs):
        self.smooth = smooth
        super().__init__(*args, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self._provides(s, T.ELEVATION, StatisticJournalType.FLOAT, U.M, None,
                       'An estimate of elevation (may come from various sources).')
        self._provides(s, T.GRADE, StatisticJournalType.FLOAT, U.PC, None,
                       'The gradient of the smoothed SRTM1 elevation.')

    def _read_dataframe(self, s, ajournal):
        from ..owners import ActivityReader
        try:
            device = StatisticJournal.at(s, ajournal.start, N.DEVICE, ActivityReader,
                                         ajournal.activity_group)
            return Statistics(s, activity_journal=ajournal, with_timespan=True). \
                by_name(ActivityReader, N.LATITUDE, N.LONGITUDE, N.DISTANCE, N.ELAPSED_TIME,
                        N.SRTM1_ELEVATION, N.ELEVATION, N.ALTITUDE,
                        N.SPHERICAL_MERCATOR_X, N.SPHERICAL_MERCATOR_Y).df, device
        except Exception as e:
            log.warning(f'Failed to generate statistics for elevation: {e}')
            raise

    def _calculate_stats(self, s, ajournal, data):
        df, device = data
        if not present(df, N.ELEVATION):
            if device and device.value == 'Edge_130' and present(df, N.ALTITUDE):
                log.info(f'Using {N.ALTITUDE} directly (barometer)')
                df[N.ELEVATION] = df[N.ALTITUDE]
                if present(df, N.SRTM1_ELEVATION): self.__improve(df)
                add_gradient(df)
            elif present(df, N.SRTM1_ELEVATION):
                df = smooth_elevation(df, smooth=self.smooth)
                # not used and may be nulls, breaking geo
                df.drop(columns=[N.ALTITUDE], errors='ignore', inplace=True)
            elif present(df, N.ALTITUDE):
                log.warning(f'Using {N.ALTITUDE} as {N.ELEVATION}')
                df[N.ELEVATION] = df[N.ALTITUDE]
                add_gradient(df)
            return add_delta_azimuth(df)
        else:
            return None

    def __improve(self, df):
        correction = df[N.SRTM1_ELEVATION].median() - df[N.ALTITUDE].median()
        log.info(f'Correcting {N.ELEVATION} by {correction:.1f}m')
        df[N.ELEVATION] += correction
        sd = sqrt((df[N.ELEVATION] - df[N.SRTM1_ELEVATION]).var())
        if sd > 10:
            log.warning(f'Large difference between measured altitude and SRTM1 data ({sd:.1f}m)')
            log.warning(f'Replacing measured altitude with SRTM1 data')
            df[N.ELEVATION] = df[N.SRTM1_ELEVATION]
        else:
            log.info(f'SD of difference between corrected altitude and SRTM1 is {sd:.1f}m')

    def _copy_results(self, s, ajournal, loader, df):
        for time, row in df.iterrows():
            if N.ELEVATION in row and not is_nan(row[N.ELEVATION]):
                loader.add_data(N.ELEVATION, ajournal, row[N.ELEVATION], time)
            if N.GRADE in row and not is_nan(row[N.GRADE]):
                loader.add_data(N.GRADE, ajournal, row[N.GRADE], time)
        self.__create_postgis(s, ajournal, df)

    def __create_postgis(self, s, ajournal, df):
        # this includes things that could have been created earlier, but if we build them from SQL
        # queries then it's significantly slower
        # skip non-essential and possibly nan values
        df = df.loc[:, [N.DISTANCE, N._delta(N.AZIMUTH), N.ELAPSED_TIME, N.LONGITUDE, N.LATITUDE, N.ELEVATION]]
        df.dropna(inplace=True)   # so all routes should be aligned (fractional positions should correspond)
        if self.__create_route(s, ajournal, df, 'route_d', N.DISTANCE):
            self.__create_centre(s, ajournal)
            self.__create_utm_srid(s, ajournal)
        self.__create_route(s, ajournal, df, 'route_a', N._delta(N.AZIMUTH))
        self.__create_route_z(s, ajournal, df, 'route_et', N.ELAPSED_TIME)
        compress_distance_time(df)
        self.__create_route_z(s, ajournal, df, 'route_edt', N.DISTANCE_TIME)

    def __create_route_z(self, s, ajournal, df, name, m):
        log.debug(f'Setting {name}')
        xyzm = list(self.__xyzm(df, m))
        line = linestringxyzm(xyzm)
        table = ajournal.__table__
        update = table.update().values(**{name: text(line)}).where(table.c.id == ajournal.id)
        # log.debug(update)
        s.execute(update)

    def __xyzm(self, df, m):
        if N.ELEVATION in df:
            for _, row in df.dropna().iterrows():
                yield row[N.LONGITUDE], row[N.LATITUDE], row[N.ELEVATION], row[m]

    def __create_route(self, s, ajournal, df, name, m):
        log.debug(f'Setting {name}')
        xym = list(self.__xym(df, m))
        line = linestringxym(xym)
        table = ajournal.__table__
        update = table.update().values(**{name: text(line)}).where(table.c.id == ajournal.id)
        # log.debug(update)
        s.execute(update)
        return bool(xym)

    def __xym(self, df, m):
        for _, row in df.dropna().iterrows():
            yield row[N.LONGITUDE], row[N.LATITUDE], row[m]

    def __create_utm_srid(self, s, ajournal):
        table = ActivityJournal.__table__
        query = select([table.c.centre]).where(table.c.id == ajournal.id)
        log.debug(query)
        row = s.execute(query).fetchone()
        lon, lat = to_shape(row[0]).coords[0]
        srid = utm_srid(lat, lon)
        update = table.update().values(utm_srid=srid).where(table.c.id == ajournal.id)
        log.debug(update)
        s.execute(update)

    def __create_centre(self, s, ajournal):
        table = ActivityJournal.__table__
        centre = f'ST_Centroid({table.c.route_d})'
        update = table.update().values(centre=text(centre)).where(table.c.id == ajournal.id)
        log.debug(update)
        s.execute(update)


def elapsed_time_to_time(df, t0):
    df[N.TIME] = pd.to_timedelta(df[N.ELAPSED_TIME], 'seconds') + t0
    return df


def compress_distance_time(df):
    # convert to m and round both to 1dp
    df[N.DISTANCE_TIME] = (df[N.DISTANCE] * 1e3 * 10 + 0.5).astype(int) * 1e7 / 10 + \
                          (df[N.ELAPSED_TIME] * 10 + 0.5).astype(int) / 10
    return df


def expand_distance_time(df):
    df[N.DISTANCE] = (df[N.DISTANCE_TIME] / 1e7).round(1)
    df[N.ELAPSED_TIME] = (df[N.DISTANCE_TIME] - df[N.DISTANCE] * 1e7).round(1)
    df[N.DISTANCE] /= 1000   # convert back to km
    return df
